"""
Functional API for dict as row
------------------------------------

"""

from typing import List, Dict, Tuple, Callable, Text, Any

from .helper import reduce_args_count, reduce_callback
from ..utils import number_format, reduce_args, reduce_kwargs



def askeys(item:Dict, excludes:List=None):
    """extract keys from item

    :param item: a dict object
    :param excludes: keys for excluded, defaults to None
    :return: a list contains keys of item


    .. doctest::  

        >>> import listorm

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.askeys(item)
        ['name', 'gender', 'age', 'location']

        >>> listorm.askeys(item, excludes=['gender', 'age'])
        ['name', 'location']

    """

    excludes = excludes or []
    return [
        key for key in item if key not in excludes
    ]



@reduce_args('keys')
def asvalues(item:Dict, keys:List, *, exact:bool=True, flat:bool=True):    
    """extract values from item that matchs the order of keys

    :param item: a dict object
    :param keys: keys for retrive values
    :param exact: if True, raise KeyError when key is not in item, defaults to True
    :param flat: When the return value is a single value, the value is returned without being included in the tuple, defaults to True
    :return: a tuple of contains values or a value


    .. doctest::

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.asvalues(item, ['name', 'gender']) #  listify keys
        ('Smith', 'M')

        >>> listorm.asvalues(item, 'name', 'gender') # also as unpacked args
        ('Smith', 'M')

        >>> listorm.asvalues(item, 'name') # unpacked when retriving single value
        'Smith'

        >>> listorm.asvalues(item, 'name', flat=False) # if flat=False, the sigle value is also packed as a tuple.
        ('Smith',)

    """

    keys = keys or askeys(item)
    result = tuple(
        item[key] for key in keys if exact or key in item
    )
    if flat and len(result) == 1:
        return result[0]
    return result


@reduce_args('keys')
def asselect(item:Dict, keys:List, *, excludes:List=None) -> Dict:
    """select key, value pair from item

    :param item: a dict object
    :param keys: the arguments for keys or a list containing keys, if nothing, returns a new dict with the same key and value as the original item.
    :param excludes: keys for excluded, defaults to None
    :return: new dict item as selected


    .. doctest::

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.asselect(item, ['name', 'gender']) # keys arg as list
        {'name': 'Smith', 'gender': 'M'}

        >>> listorm.asselect(item, 'name', 'gender') # equal from above that keys as unpacked args
        {'name': 'Smith', 'gender': 'M'}

        >>> listorm.asselect(item) # pass a nothing of keys arg
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.asselect(item, excludes=['gender']) # useful when need to drop item
        {'name': 'Smith', 'age': 17, 'location': 'USA'}

    """
    keys = keys or askeys(item)
    keys = [
        key for key in keys
        if key not in (excludes or [])
    ]
    return {
        key:item[key] for key in keys if key in item
    }


@reduce_kwargs('keymap')
def addkeys(item:Dict, *, keymap:Dict) -> Dict:
    """extends item keys values via value or callback

    :param item: a dict object
    :param keymap: key: value pair dict of items to be added
    :return: Expanded existing items


    .. doctest::

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17}

        >>> listorm.addkeys(item, keymap={'location': 'USA'})
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.addkeys(item, location='USA') # also keymap can converted into kwargs style
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        
        # add key value via function
        >>> listorm.addkeys(item,
        ...    gender_age=lambda gender, age: "{}/{}".format(gender, age)
        ... )
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'gender_age': 'M/17'}


    .. note::
        * The lambda callback function used in this library is applied according to the number and name of the argument 
        * Refers to the keys in the item by specifying it as the lambda arguments


        .. code-block:: python

            lambda gender, age: "{}/{}".format(gender, age) # gender and age are key name of item
        * For refer to all items in a lambda, you can use **kwargs pattern, for example, when reffering space character exists in item

        .. code-block:: python

            lambda **kwargs: kwargs['age of ultron'] == 19  # (ex: when space exists in key name)

    """
    
    added = {}
    for key, app in keymap.items():
        added.update(reduce_callback(item, key, app))

    item = dict(item)
    item.update(added)
    return item


@reduce_kwargs('renamemap')
def asrename(item:Dict, *, renamemap:Dict) -> Dict:
    """change key as to another name

    :param item: a dict object
    :param renamemap: old: new name pair of dict
    :return: newly renamed dict


    .. doctest::

        >>> import listorm

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> listorm.asrename(item, renamemap={'name': 'first_name', 'gender': 'sex', 'location': 'country'})
        {'first_name': 'Smith', 'sex': 'M', 'age': 17, 'country': 'USA'}

        >>> listorm.asrename(item, name='first_name', gender='sex', location='country') # also as uppacked kwarg style
        {'first_name': 'Smith', 'sex': 'M', 'age': 17, 'country': 'USA'}

    """
    return {
        renamemap.get(key, key): value
        for key, value in item.items()
    }


@reduce_kwargs('defaultmap')
def asdefault(item:Dict, *, defaultmap:Dict) -> Dict:
    """fill values from defaults if key not in existing item

    :param item: a dict object
    :param defaultmap: defaults key:value mapping for missing items
    :return: newly normalized dict


    .. doctest::
    
        >>> normal_item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> # nothing is happend when applied at normal_item which has not missing values at location
        >>> listorm.asdefault(normal_item, location='Unknown')
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> # item that has missing keys for location and age
        >>> missing_location_and_age = {'name': 'Smith', 'gender': 'M'}

        >>> # filled with defaults values at missing keys
        >>> listorm.asdefault(missing_location_and_age, location='Earth', age=10)
        {'name': 'Smith', 'gender': 'M', 'location': 'Earth', 'age': 10}

        >>> # fill value by function
        >>> missing_location = {'name': 'Smith', 'gender': 'M', 'age': 17}

        >>> def fill_location(age): # argument name age matched with age as key in item
        ...     return 'Universe' if age > 100 else 'Earth'

        >>> # Changing the location by referring to the neighboring key age
        >>> listorm.asdefault(missing_location, location=fill_location)
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'Earth'}

    """

    defaults = asselect(defaultmap, excludes=askeys(item))
    return addkeys(item, keymap=defaults)


def asdiff(item1:Dict, item2:Dict) -> List:
    """Finds keys with different value from common keys of two items

    :param item1: a dict object
    :param item2: anothor dict object
    :return: list of keys 


    .. doctest::

        >>> item1 = {'name': 'Xiaomi', 'product': 'battery', 'amount':7}
        >>> item2 = {'name': 'Park', 'product': 'battery', 'amount':2}
        >>> listorm.asdiff(item1, item2)
        ['name', 'amount']

    """
    return [
        key1
        for key1, value1 in item1.items()
        for key2, value2 in item2.items()
        if key1 == key2 and value1 != value2
    ]


def asnumformat(dict:Dict, examples:Dict):
    return {
        key: number_format(value, examples.get(key))
        for key, value in dict.items()
    }


@reduce_kwargs('keymap')
def asmap(item:Dict, keymap:Dict):
    applied = {} 
    for key, value in item.items():
        app = keymap.get(key)
        if not app:
            applied[key] = value
            continue
        app = keymap[key]
        applied[key] = reduce_args_count(app, value, item)
    return applied


@reduce_kwargs('updatemap')
def asupdate(item:Dict, updatemap:Dict):
    """update item values

    :param item: a dict object
    :param updatemap: key:value|callable mapping for update values
    :return: updated item


    .. doctest::

        >>> item = {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'usa'}

        >>> # update by assign simple value directly
        >>> listorm.asupdate(item, age=100) 
        {'name': 'Smith', 'gender': 'M', 'age': 100, 'location': 'usa'}

        >>> # update value via function
        >>> listorm.asupdate(item, location=lambda location: str.upper(location))
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        # as shortcut instead of using lambda
        >>> listorm.asupdate(item, location=str.upper)
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

    """
            
    updated = {}
    for key, value in item.items():
        app = updatemap.get(key)
        if not app:
            updated[key] = value
        else:
            updated.update(reduce_callback(item, key, app))        
    return updated



from io import TextIOBase, BufferedReader, StringIO, BytesIO



def get_stringio(*args):
    return StringIO(*args)


def get_bytesio(*args):
    return BytesIO(*args)


def reduce_csv_input(file, **kwargs):
    if isinstance(file, str):
        with open(file, **kwargs) as fp:
            content = fp.read()
            return get_stringio(content)
    elif isinstance(file, TextIOBase):
        return file
    raise ValueError('Invalid csv file type')


def reduce_excel_input(file):
    if isinstance(file, str):
        with open(file, 'rb') as fp:
            return BytesIO(fp.read())
    elif isinstance(file, bytes):
        return BytesIO(file)
    elif isinstance(file, BufferedReader):
        return file
    else:
        raise ValueError('Invalid excel file type')




from typing import List, Dict, Text

import openpyxl


from .io import get_bytesio, reduce_excel_input
from ..records import fillmissed, values, askeys, select



def _find_table(values, fields_contains:List=None, start_rows:int=None, start_cols:int=None):
    def _find_start_row_by_fields(values, fields):
        fieldset = set(fields)
        for i, row in enumerate(values):
            if fieldset < set(row):
                return i
        raise ValueError("Cannot find start row number by {}".format(fields))

    def _find_column_offset(row):
        for i, col in enumerate(row):
            if not col:
                continue
            return i

    def _slice_padding(values, nrow, ncol):
        values = values[nrow:]
        return list(map(lambda row: row[ncol:], values))

    values = [val for val in values]
    if start_rows is None:
        start_rows = _find_start_row_by_fields(values, fields_contains)
    if start_cols is None:
        start_cols = _find_column_offset(values[start_rows])

    if start_rows == 0 and start_cols == 0:
        return values
    return _slice_padding(values, start_rows, start_cols)


def _normalize_columns(fields):
    results = []
    trantab = {
        '\n': '',
        '_x000D_': '',
        '_x000d_': '',
    }
    for col in fields:
        for t, c in trantab.items():
            col = col.replace(t, c)
        col = col.strip()
        results.append(col)
    return results


def read_excel(file=None, search_fields:List=None, sheet_name:Text=None, start_rows:int=0, start_cols:int=0, prettify_column_names:bool=True):
    '''Excel File or byte Content of Excel to Listorm object
    '''
    file = reduce_excel_input(file)
    workbook= openpyxl.load_workbook(file)
    worksheet = workbook[sheet_name] if sheet_name else workbook.active
    if search_fields:
        table = _find_table(worksheet.values, search_fields)
    else:
        table = _find_table(worksheet.values, start_rows=start_rows, start_cols=start_cols)
    workbook.close()
    fields, *records = table
    if prettify_column_names:
        fields = _normalize_columns(fields)
    return [
        dict(zip(fields, row)) for row in records
    ]


def write_excel(records:List[Dict], filename=None, fill_miss=True):
    if not records:
        raise ValueError('Cannot write excel from Empty list')
    if fill_miss:
        records = fillmissed(records)

    output = get_bytesio()
    workbook = openpyxl.Workbook()
    worksheet = workbook.active
    fields = askeys(records[0])
    selected = select(records, fields)
    rows = values(selected, flat_one=False)
    worksheet.append(fields)
    for row in rows:
        worksheet.append(row)

    if not filename:
        workbook.save(output)
        content = output.getvalue()
        return content
    workbook.save(filename)



'''
Api for dict and list
=====================

Basic Concepts
--------------

* To maintain immutability, create new objects without altering existing ones
* When transforming a dict or list, try to keep the original order as much as possible, since newer versions of python, the key order of the dict is available




'''
from .records import *
from .extensions import read_excel, read_csv, write_csv, write_excel


import csv, xlrd, xlsxwriter
from io import TextIOBase, BufferedReader, StringIO, BytesIO


def to_io(path, *args, **kwargs):
    if isinstance(path, str):
        return open(path, *args, **kwargs)
    elif isinstance(path, (TextIOBase, BufferedReader)):
        return path
    else:
        raise ValueError('Either file path or fp must be specified')


def read_csv(file, encoding='utf-8'):
    fp = to_io(file, encoding=encoding)
    csv_reader = csv.reader(fp)
    fields = next(csv_reader)
    result = [dict(zip(fields, map(str, row))) for row in csv_reader]
    return result
    

def read_excel(file, sheet_index=0, starts_row=0):
    '''Excel File or byte Content of Excel to Listorm object
    '''
    fp = to_io(file, 'rb')
    content = fp.read()
    wb = xlrd.open_workbook(file_contents=content)
    ws = wb.sheet_by_index(sheet_index)
    fields = ws.row_values(starts_row)
    result = [dict(zip(fields, map(str, ws.row_values(r)))) for r in range(starts_row+1, ws.nrows)]
    return result
  


def write_csv(records:list[dict], filename=None):
    output = StringIO()
    writer = csv.DictWriter(output, lineterminator='\n')
    writer.writeheader()
    for row in records:
        writer.writerow(row)
    if filename:
        with open(filename, 'w') as fp:
            fp.write(output.getvalue())
    else:
        return output.getvalue()


def to_excel(records:list[dict], filename=None):
    '''If filnames is None, returns filecontents
    '''
    output = BytesIO()
    wb = xlsxwriter.Workbook(output)
    ws = wb.add_worksheet()

    for i, row in enumerate(records):
        if i == 0:
            values = [k for k in row]
        else:
            values = [v for k, v in row.items()]
        ws.write_row(i,0, values)
    wb.close()

    if filename:
        with open(filename, 'wb') as fp:
            fp.write(output.getvalue())
    else:
        return output.getvalue()


import inspect, functools



def get_param_type(func, param_name):
    sig = inspect.signature(func)
    for pname, param in sig.parameters.items():
        if pname == param_name:
            return param.kind
    raise TypeError('Cannot found {} in parameters of {}'.format(param_name, func))


def get_param_index(func, param_name):
    sig = inspect.signature(func)
    params = list(sig.parameters)
    try:
        index = params.index(param_name)
    except ValueError:
        raise ValueError("'{}' is not in parameters of {}".format(param_name, func))
    return index


def get_default(func, param_name):
    sig = inspect.signature(func)
    param = sig.parameters[param_name]
    return param.default


def update_params(func, param_name, callback, *args, **kwargs):
    args = list(args)
    kwargs = dict(kwargs)
    index = get_param_index(func, param_name)
    if index < len(args):
        args[index] = callback(args[index])
    else:
        if param_name in kwargs:
            kwargs[param_name] = callback(kwargs[param_name])
        else:
            kwargs[param_name] = get_default(func, param_name)

    return tuple(args), kwargs


def reduce_args(target):
    def wrapper(func):
        @functools.wraps(func)
        def reduce(*args, **kwargs):
            if target in kwargs:
                return func(*args, **kwargs)
            index = get_param_index(func, target)
            head, tail = args[:index], args[index:]
            reduced = []
            for arg in tail:
                if isinstance(arg, (list, tuple)):
                    reduced += arg
                else:
                    reduced.append(arg)
            return func(*head, reduced, **kwargs)
        return reduce
    return wrapper


def reduce_kwargs(target):
    def wrapper(func):
        @functools.wraps(func)
        def reduce(*args, **kwargs):
            if target in kwargs:
                return func(*args, **kwargs)            
            sig = inspect.signature(func)
            kw = {}
            rest = {}
            for key, value in kwargs.items():
                if key not in sig.parameters:
                    kw[key] = value
                else:
                    rest[key] = value
            kwargs = {
                target: kw
            }
            kwargs.update(rest)
            return func(*args, **kwargs)
        return reduce
    return wrapper
            

def _tuplizer(value):
    if value is None:
        return value
    if isinstance(value, (list, tuple)):
        return value
    return value,

def pluralize_params(*targets):
    targets = list(targets)
    def wrapper(func):
        @functools.wraps(func)
        def plural(*args, **kwargs):
            while targets:
                target = targets.pop()
                args, kwargs = update_params(
                    func, target,
                    _tuplizer,
                    *args, **kwargs
                )
            return func(*args, **kwargs)
        return plural
    return wrapper


def filer_kwargs(item, func):
    sig = inspect.signature(func)
    return {
        p: item[p] for p in sig.parameters if p in item
    }



'''
test doc string
'''


class UniqueConstraintError(Exception):
    '''Unique constratint failed'''
    pass

class JoinKeyDoesNotExists(Exception):
    '''join key does not exists'''
    pass

class ApplyFunctionArgumentCountError(Exception):

    def __str__(self):
        func, *_ = self.args
        msg = "Arguments count for applying function({}) must be 1 or 2".format(func.__name__)
        return msg



from ..api import extensions
from .orm import Listorm



def read_excel(file=None, fields_contains:list=None, uniques=None, sheet_name=None, start_rows=0, start_cols=0, prettify_column_names=True):
    '''import records from excel

    :param file: filename|fp, defaults to None
    :param fields_contains: a parts of column names to find excel table, defaults to None
    :param uniques: unique keys, defaults to None
    :param sheet_name: sheetname where records exists, defaults to None
    :param start_rows: cell rows where table starts, defaults to 0
    :param start_cols: cell col index where table starts, defaults to 0
    :param prettify_column_names: if True, remove seps linebreaker space..etc , defaults to True
    :return: Listorm object


    .. note::
        **fields_contains** : It finds a table that exists at any location in the Excel sheet through a part of the field name

    '''
    records = extensions.read_excel(
        file, fields_contains, sheet_name, start_rows, start_cols,
        prettify_column_names=prettify_column_names
    )
    return Listorm(
        records, fill_missed=False,
        uniques=uniques,
    )



def read_csv(file, encoding='utf-8', uniques=None, **csv_reader_kwargs):
    '''import records from csv file

    :param file: filename|fp
    :param encoding: encoding of csv text source, defaults to 'utf-8'
    :param uniques: columns to check unique, defaults to None
    :param csv_reader_kwargs: kwargs of csv.reader which builtin module csv
    :return: Listorm object
    '''

    records = extensions.read_csv(file, encoding=encoding, **csv_reader_kwargs)
    return Listorm(
        records, fill_missed=False,
        uniques=uniques
    )